/* Copyright 2015 Corentin Noël <corentin@elementary.io>
*
* This file is part of Cafeteria
*
* Cafeteria is free software: you can redistribute it
* and/or modify it under the terms of the GNU General Public License as
* published by the Free Software Foundation, either version 3 of the
* License, or (at your option) any later version.
*
* Cafeteria is distributed in the hope that it will be
* useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
* Public License for more details.
*
* You should have received a copy of the GNU General Public License along
* with Cafeteria. If not, see http://www.gnu.org/licenses/.
*/

public class Cafeteria.CartView : Gtk.Grid {
    public signal void product_number (bool sensitive);

    private Gtk.ListBox list_box;
    private Gtk.Label total_label;
    private Gtk.Label total_premium_label;

    private uint price = 0;
    private uint price_member = 0;

    public CartView () {
        
    }

    construct {
        margin = 12;
        row_spacing = 6;
        orientation = Gtk.Orientation.VERTICAL;
        list_box = new Gtk.ListBox ();
        list_box.selection_mode = Gtk.SelectionMode.NONE;
        var scrolledwindow = new Gtk.ScrolledWindow (null, null);
        scrolledwindow.add (list_box);
        scrolledwindow.expand = true;
        add (scrolledwindow);

        var total_desc_label = new Gtk.Label (_("Total:"));
        ((Gtk.Misc) total_desc_label).xalign = 1;
        total_label = new Gtk.Label (Utils.format_money (0U));
        total_label.margin_start = 6;
        ((Gtk.Misc) total_label).xalign = 0;
        var total_premium_desc_label = new Gtk.Label (_("Total Premium:"));
        ((Gtk.Misc) total_premium_desc_label).xalign = 1;
        total_premium_label = new Gtk.Label (Utils.format_money (0U));
        total_premium_label.margin_start = 6;
        ((Gtk.Misc) total_premium_label).xalign = 0;

        var bottom_grid = new Gtk.Grid ();
        bottom_grid.halign = Gtk.Align.END;
        bottom_grid.column_spacing = 12;
        bottom_grid.column_homogeneous = true;
        bottom_grid.attach (total_desc_label, 0, 0, 1, 1);
        bottom_grid.attach (total_label, 1, 0, 1, 1);
        bottom_grid.attach (total_premium_desc_label, 0, 1, 1, 1);
        bottom_grid.attach (total_premium_label, 1, 1, 1, 1);
        add (bottom_grid);

        var info_label = new Gtk.Label (_("Start by adding a product"));
        info_label.wrap = true;
        info_label.justification = Gtk.Justification.CENTER;
        info_label.get_style_context ().add_class ("h1");
        info_label.halign = Gtk.Align.CENTER;
        info_label.valign = Gtk.Align.CENTER;
        info_label.show_all ();
        list_box.set_placeholder (info_label);
    }

    public void add_product (Product product) {
        var children = list_box.get_children ();
        for (uint i = 0; i < children.length (); i++) {
            var row = children.nth_data (i) as ProductRow;
            if (row.product == product) {
                row.number++;
                product_number_changed ();
                return;
            }
        }

        var row = new ProductRow (product);
        row.show_all ();
        row.notify["number"].connect (() => product_number_changed ());
        row.destroy.connect (() => product_number_changed ());
        list_box.add (row);
        product_number_changed ();
    }

    private void product_number_changed () {
        var number = get_product_number ();
        product_number (number != 0);
        price = 0;
        price_member = 0;
        list_box.get_children ().foreach ((child) => {
            var row = child as ProductRow;
            if (row != null) {
                price += row.product.price * row.number;
                price_member += row.product.price_member * row.number;
            }
        });

        total_label.label = Utils.format_money (price);
        total_premium_label.label = Utils.format_money (price_member);
    }

    public uint get_product_number () {
        uint i = 0;
        list_box.get_children ().foreach ((child) => {
            var row = child as ProductRow;
            if (row != null) {
                i += row.number;
            }
        });

        return i;
    }

    public void paid () {
        unowned Database db = Database.get_default ();
        list_box.get_children ().foreach ((child) => {
            var row = child as ProductRow;
            if (row != null) {
                var product = row.product;
                if (product.quantity != -1) {
                    db.save_stock.begin (product, product.quantity - (int)row.number);
                }
            }
        });

        clear ();
    }

    public void clear () {
        list_box.get_children ().foreach ((child) => {
            child.destroy ();
        });
    }

    public uint get_price (bool member) {
        if (member) {
            return price_member;
        } else {
            return price;
        }
    }

    public class ProductRow : Gtk.ListBoxRow {
        public uint number {
            public get {
                return (uint) number_spin.value;
            }
            public set {
                number_spin.@value = (uint)value;
            }
        }
        public Product product;
        Gtk.Label name_label;
        Gtk.Button delete_button;
        Gtk.SpinButton number_spin;
        public ProductRow (Product product) {
            this.product = product;
            int max = product.quantity >= 0 ? int.min (15, product.quantity) : 15;
            number_spin.set_range (1, max);
            name_label.label = product.name;
        }

        construct {
            activatable = false;
            selectable = false;
            var grid = new Gtk.Grid ();
            grid.orientation = Gtk.Orientation.HORIZONTAL;
            grid.margin = 6;
            grid.column_spacing = 12;
            name_label = new Gtk.Label (null);
            ((Gtk.Misc) name_label).xalign = 0;
            name_label.hexpand = true;
            number_spin = new Gtk.SpinButton.with_range (1, 15, 1);
            number_spin.value = 1;
            number_spin.value_changed.connect (() => {
                notify_property ("number");
            });

            delete_button = new Gtk.Button.from_icon_name ("edit-delete-symbolic", Gtk.IconSize.BUTTON);
            delete_button.get_style_context ().add_class (Gtk.STYLE_CLASS_FLAT);
            delete_button.get_style_context ().add_class (Gtk.STYLE_CLASS_DESTRUCTIVE_ACTION);
            delete_button.clicked.connect (() => {
                destroy ();
            });
            grid.add (delete_button);
            grid.add (name_label);
            grid.add (number_spin);
            add (grid);
        }
    }
}
