/* Copyright 2015 Corentin Noël <corentin@elementary.io>
*
* This file is part of Cafeteria
*
* Cafeteria is free software: you can redistribute it
* and/or modify it under the terms of the GNU General Public License as
* published by the Free Software Foundation, either version 3 of the
* License, or (at your option) any later version.
*
* Cafeteria is distributed in the hope that it will be
* useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
* Public License for more details.
*
* You should have received a copy of the GNU General Public License along
* with Cafeteria. If not, see http://www.gnu.org/licenses/.
*/

public class Cafeteria.Database : GLib.Object {
    private static Database database;
    public static unowned Database get_default () {
        if (database == null) {
            database = new Database ();
        }

        return database;
    }

    public signal void new_product (Product product);

    public bool locked {
        get {
            return server_token == null;
        }
    }

    private Gee.TreeMap<int, Product> products;
    private const string API_HOST = "http://webteam-beta.asso-ensea.fr/app_dev.php/api/kfet/v1/";
    private const string CLIENT_UID = "12rjvvl7va7ks4wgk04cc88o4w0k8g8c4kc4koo444cooccgs4";
    private string server_token = null;
    private DateTime token_expiration = null;
    private Soup.Session session;

    private Database () {
        
    }

    construct {
        products = new Gee.TreeMap<int, Product> ();
        session = new Soup.Session ();
    }

    public async bool identify_server (string server_card_id) {
        try {
            var uri = new Soup.URI (API_HOST + "token");
            uri.set_query_from_fields ("card_uid", server_card_id, "client_uid", CLIENT_UID);
            var message = new Soup.Message.from_uri ("GET", uri);
            var steam = yield session.send_async (message);
            if (message.status_code != Soup.Status.OK) {
                critical (uri.to_string (false));
                debug ("%s", (string) message.response_body.data);
                return false;
            }

            var parser = new Json.Parser ();
            yield parser.load_from_stream_async (steam);
            weak Json.Object root_object = parser.get_root ().get_object ();
            server_token = root_object.get_string_member ("token");
            token_expiration = new DateTime.from_unix_local (root_object.get_int_member ("expire_at"));
            notify_property ("locked");
            refresh_products.begin ();
            return true;
        } catch (Error e) {
            critical (e.message);
            return false;
        }
    }

    public void log_out () {
        server_token = null;
        token_expiration = null;
        notify_property ("locked");
    }

    public Gee.TreeSet<Product> get_products () {
        var treeset = new Gee.TreeSet<Product> ();
        treeset.add_all (products.values);
        return treeset;
    }

    public async void refresh_products () {
        if (server_token == null) {
            return;
        }

        try {
            var uri = new Soup.URI (API_HOST + "product/list");
            uri.set_query_from_fields ("token", server_token);
            var message = new Soup.Message.from_uri ("GET", uri);
            var steam = yield session.send_async (message);
            if (message.status_code != Soup.Status.OK) {
                debug ("%s", (string) message.response_body.data);
                return;
            }

            var parser = new Json.Parser ();
            yield parser.load_from_stream_async (steam);
            var old_products_list = new Gee.TreeSet<Product> ();
            old_products_list.add_all (products.values);
            weak Json.Array root_array = parser.get_root ().get_array ();
            var products_list = new Gee.TreeSet<Product> ();
            root_array.get_elements ().foreach ((node) => products_list.add (parse_product_node (node)));
            foreach (var product in products_list) {
                if (old_products_list.contains (product)) {
                    old_products_list.remove (product);
                } else {
                    new_product (product);
                }
            }

            // For now we just disable the remaining products
            foreach (var old_product in old_products_list) {
                old_product.quantity = 0;
            }
        } catch (Error e) {
            critical (e.message);
        }
    }

    private Product parse_product_node (Json.Node node, Product? parent = null) {
        weak Json.Object objet = node.get_object ();
        var id = (int)objet.get_int_member ("id");
        Product product = null;
        product = products.get (id);
        if (product == null) {
            product = new Product ();
            product.SKU = id;
            product.name = objet.get_string_member ("name");
            product.category = Product.Category.from_string (objet.get_string_member ("category"));
            product.menu_activator = objet.get_boolean_member ("menu_activator");
            products.set (id, product);
        }

        if (product.parent != parent) {
            if (product.parent != null) {
                product.parent.children.remove (product);
                product.parent.children_changed (product, false);
            }

            product.parent = parent;
            if (parent != null) {
                parent.children.add (product);
                parent.children_changed (product, true);
            }
        }
        product.price = (uint)objet.get_int_member ("price");
        product.price_member = (uint)objet.get_int_member ("price_member");
        product.price_menu = (uint)objet.get_int_member ("price_menu");
        product.price_member_menu = (uint)objet.get_int_member ("price_member_menu");
        product.quantity = (int)objet.get_int_member ("quantity");
        weak Json.Array variants = objet.get_array_member ("variants");
        variants.get_elements ().foreach ((node) => parse_product_node (node, product));
        return product;
    }

    public async UserInfo? get_user_from_device (string device_id) {
        if (server_token == null) {
            return null;
        }

        try {
            var uri = new Soup.URI (API_HOST + "user/get_from_card");
            uri.set_query_from_fields ("token", server_token, "card_uid", device_id);
            var message = new Soup.Message.from_uri ("GET", uri);
            var steam = yield session.send_async (message);
            if (message.status_code != Soup.Status.OK) {
                debug ("%s", (string) message.response_body.data);
                return null;
            }

            var parser = new Json.Parser ();
            yield parser.load_from_stream_async (steam);
            weak Json.Object root = parser.get_root ().get_object ();
            var info = new UserInfo ();
            info.id = root.get_int_member ("id");
            info.name = root.get_string_member ("name");
            info.first_name = root.get_string_member ("first_name");
            info.money = (uint)root.get_int_member ("money");
            info.member = root.get_boolean_member ("member");
            return info;
        } catch (Error e) {
            critical (e.message);
            return null;
        }
    }

    public async bool user_pay (int64 user_id, int amount) {
        if (server_token == null) {
            return false;
        }

        try {
            var uri = new Soup.URI (API_HOST + "user/%lld/change_money".printf (user_id));
            uri.set_query_from_fields ("token", server_token, "money_delta", "%d".printf (-amount));
            var message = new Soup.Message.from_uri ("GET", uri);
            yield session.send_async (message);
            if (message.status_code != Soup.Status.OK) {
                debug ("%s", (string) message.response_body.data);
                return false;
            }

            return true;
        } catch (Error e) {
            critical (e.message);
            return false;
        }
    }

    public async void user_refill (int64 user_id, int amount) {
        if (server_token == null) {
            return;
        }

        try {
            var uri = new Soup.URI (API_HOST + "user/%lld/change_money".printf (user_id));
            uri.set_query_from_fields ("token", server_token, "money_delta", "%d".printf (amount));
            var message = new Soup.Message.from_uri ("GET", uri);
            yield session.send_async (message);
            if (message.status_code != Soup.Status.OK) {
                debug ("%s", (string) message.response_body.data);
                return;
            }

        } catch (Error e) {
            critical (e.message);
            return;
        }
    }

    public async Gee.TreeSet<UserInfo>? get_users_for_promo (uint promo) {
        if (server_token == null) {
            return null;
        }

        try {
            var uri = new Soup.URI (API_HOST + "user/list");
            uri.set_query_from_fields ("token", server_token, "promo", "%u".printf (promo));
            var message = new Soup.Message.from_uri ("GET", uri);
            var steam = yield session.send_async (message);
            if (message.status_code != Soup.Status.OK) {
                debug ("%s", (string) message.response_body.data);
                return null;
            }

            var parser = new Json.Parser ();
            yield parser.load_from_stream_async (steam);
            weak Json.Array root_array = parser.get_root ().get_array ();
            var users = new Gee.TreeSet<UserInfo> ();
            root_array.get_elements ().foreach ((node) => {
                weak Json.Object root = node.get_object ();
                var info = new UserInfo ();
                info.id = root.get_int_member ("id");
                info.name = root.get_string_member ("name");
                info.first_name = root.get_string_member ("first_name");
                info.member = root.get_boolean_member ("member");
                users.add (info);
            });

            return users;
        } catch (Error e) {
            critical (e.message);
            return null;
        }
    }

    public async bool save_stock (Product product, int new_stock) {
        if (server_token == null) {
            return false;
        }

        try {
            var uri = new Soup.URI (API_HOST + "product/%d/alterate_quantity".printf (product.SKU));
            uri.set_query_from_fields ("token", server_token, "quantity_delta", "%d".printf (new_stock - product.quantity));
            var message = new Soup.Message.from_uri ("GET", uri);
            yield session.send_async (message);
            if (message.status_code != Soup.Status.OK) {
                debug ("%s", (string) message.response_body.data);
                return false;
            }

            product.quantity = new_stock;
            return true;
        } catch (Error e) {
            critical (e.message);
            return false;
        }
    }

    public async bool add_card (int64 user_id, string card_uid, string card_name) {
        if (server_token == null) {
            return false;
        }

        try {
            var uri = new Soup.URI (API_HOST + "user/%lld/card".printf (user_id));
            uri.set_query_from_fields ("token", server_token, "card_uid", card_uid, "name", card_name);
            var message = new Soup.Message.from_uri ("POST", uri);
            yield session.send_async (message);
            if (message.status_code != Soup.Status.OK) {
                debug ("%s", (string) message.response_body.data);
                return false;
            }

            return true;
        } catch (Error e) {
            critical (e.message);
            return false;
        }
    }

    public async void remove_card (int64 user_id, string card_uid) {
        if (server_token == null) {
            return;
        }

        try {
            var uri = new Soup.URI (API_HOST + "user/%lld/card".printf (user_id));
            uri.set_query_from_fields ("token", server_token, "card_uid", card_uid);
            var message = new Soup.Message.from_uri ("DELETE", uri);
            yield session.send_async (message);
            if (message.status_code != Soup.Status.OK) {
                debug ("%s", (string) message.response_body.data);
                return;
            }
        } catch (Error e) {
            critical (e.message);
            return;
        }
    }

    public async Gee.TreeSet<CardInfo>? get_cards (int64 user_id) {
        if (server_token == null) {
            return null;
        }

        try {
            var uri = new Soup.URI (API_HOST + "user/%lld/card".printf (user_id));
            uri.set_query_from_fields ("token", server_token);
            var message = new Soup.Message.from_uri ("GET", uri);
            var steam = yield session.send_async (message);
            if (message.status_code != Soup.Status.OK) {
                debug ("%s", (string) message.response_body.data);
                return null;
            }


            var parser = new Json.Parser ();
            yield parser.load_from_stream_async (steam);
            weak Json.Array root_array = parser.get_root ().get_array ();
            var cards = new Gee.TreeSet<CardInfo> ();
            root_array.get_elements ().foreach ((node) => {
                weak Json.Object root = node.get_object ();
                var info = new CardInfo ();
                info.uid = root.get_string_member ("uid");
                info.name = root.get_string_member ("name");
                cards.add (info);
            });

            return cards;
        } catch (Error e) {
            critical (e.message);
            return null;
        }
    }
}
